import queue
from multiprocessing.managers import BaseManager

from kindenty.base.ExceptionModel import ArgsException


class Data:
    keys = []
    values = []

    def put(self, k, v):
        self.keys.append(k)
        self.values.append(v)

    def get(self, k):
        if not isinstance(k, (str, int)):
            raise ArgsException('参数错误')
        if isinstance(k, str):
            i = self.keys.index(k)
            return self.values[i]
        if isinstance(k, int):
            return self.values[k]

    def remove(self, k):
        i = self.keys.index(k)
        self.values.pop(i)
        self.keys.remove(k)

    def keys(self):
        return self.keys

    def values(self):
        return self.values


class Master(BaseManager):
    workerRegisterListQueue = queue.Queue()
    realMarketDataQueue = {}
    finishSharesCodeQueue = queue.Queue()
    submitOrderQueue = queue.Queue()
    cancelOrderQueue = queue.Queue()
    testMainOrderQueue = queue.Queue()
    sharesSubscribeQueue = queue.Queue()
    unSharesSubscribeQueue = queue.Queue()

    def __init__(self, ip: bytes = b'127.0.0.1', port: int = 5000, password: bytes = b'123456'):
        Master.register('getSubmitOrder', callable=getSubmitOrderQueue)
        Master.register('getCancelOrder', callable=getCancelOrderQueue)
        Master.register('getRealMarketData', callable=getRealMarketDataQueue)
        Master.register('workerRegisterList', callable=getWorkerRegisterList)
        Master.register('getFinishSharesCode', callable=getFinishSharesCodeQueue)
        Master.register('realMarketData', callable=getRealMarketData)
        Master.register('getTestMainOrder', callable=getTestMainOrderQueue)
        Master.register('getSharesSubscribe', callable=getSharesSubscribeQueue)
        Master.register('getUnSharesSubscribe', callable=getUnSharesSubscribeQueue)
        self.registerRealMarket = dict()
        super().__init__(address=(ip, port), authkey=password)

    def getRegisterRealMarketQueue(self, key):
        if key not in Master.realMarketDataQueue.keys():
            queue = self.getRealMarketData(key)
            self.registerRealMarket[key] = queue
        return self.registerRealMarket.get(key)


def getTestMainOrderQueue():
    return Master.testMainOrderQueue


def getRealMarketData():
    return Master.realMarketDataQueue


def getRealMarketDataQueue(sharesCode):
    if sharesCode not in Master.realMarketDataQueue.keys():
        Master.realMarketDataQueue[sharesCode] = queue.Queue()
    return Master.realMarketDataQueue[sharesCode]


def getFinishSharesCodeQueue():
    return Master.finishSharesCodeQueue


def getWorkerRegisterList():
    return Master.workerRegisterListQueue


def getSubmitOrderQueue():
    return Master.submitOrderQueue


def getCancelOrderQueue():
    return Master.cancelOrderQueue


def getSharesSubscribeQueue():
    return Master.sharesSubscribeQueue


def getUnSharesSubscribeQueue():
    return Master.unSharesSubscribeQueue


class Worker(BaseManager):
    def __init__(self, ip: bytes = b'127.0.0.1', port: int = 5000, password: bytes = b'123456'):
        Worker.register('getSubmitOrder')
        Worker.register('getRealMarketData')
        Worker.register('getCancelOrder')
        Worker.register('workerRegisterList')
        Worker.register('getFinishSharesCode')
        Worker.register('getTestMainOrder')
        Worker.register('getSharesSubscribe')
        Worker.register('getUnSharesSubscribe')
        self.registerRealMarket = dict()
        super().__init__(address=(ip, port), authkey=password)

    def getRegisterRealMarketQueue(self, key):
        if key not in Master.realMarketDataQueue.keys():
            queue = self.getRealMarketData(key)
            self.registerRealMarket[key] = queue
        return self.registerRealMarket.get(key)



from enum import Enum


class OperationTypeEnum(Enum):
    add = 'add'
    update = 'update'
    disable = 'disable'
    stopSell = 'stopSell'
    startSell = 'startSell'


class HistoryOrderUpdateEnum(Enum):
    history = 'history'


class CategoryEnum(Enum):

    credit = 'credit'  # 信用账户
    shares = 'shares'  # 股票账户


class TradeSide(Enum):

    BUY = '1'
    MARGIN_BUY = 'A'
    DEBIT_BUY = 'C'
    SELL = '2'
    MARGIN_SELL = 'B'
    DEBIT_SELL = 'D'


class PriceType(Enum):
    STOCK_LIMIT = '0'

if __name__ == '__main__':
    side = TradeSide('1')
    print(side)
    side2 = TradeSide['BUY']
    print(side2)



from kindenty.base.BaseDao import BaseDao, UpdateMethod, QueryMethod, SelectMethod, \
    transaction
from kindenty.model.DataDto import AccountTradeLogDto, RealMarketData, YieldDataDto, \
    AssistDto, UserInvestmentDto, SellStatusDto
from kindenty.model.DatabaseModel import Account, AccountPosition, RevokeOrderLog, \
    SharesQuotation, Strategy, TradeLog, OpenTradeLog, WorkerRegister, HistoryData, SharesQuotationHistory, \
    TradeLogTest, YieldLog, AccountAssetsLog, ContrastLineData, LoginUser, UserInvestmentLog


class AccountDao(BaseDao):
    __mode__ = Account

    deleteByAccount = UpdateMethod("delete from account where account=':account'")

    updateAccount = UpdateMethod(
        "update account set account_type=':accountType', category=':category',priority=':priority' where account=':account'")

    getByAccount = SelectMethod(
        "select id,account,category,account_type as accountType,balance,priority from account where account=':account' ",
        Account)

    getById = SelectMethod("select id,account,category,account_type as accountType,balance,priority from account where id=:id ",
                           Account)

    getAccountByTop1 = SelectMethod(
        "select id,category,account,account_type as accountType,balance,priority from account limit 1",
        Account)

    getAccount = SelectMethod(
        "select id,category,account,account_type as accountType,balance,priority from account where account=':account' and category=':category' and account_type=':accountType' ",
        Account)
    queryByIds = QueryMethod(
        "select id,category,account,account_type as accountType, balance,priority from account where id in (:ids) ",
        Account)

    queryCreditBySharesCode = QueryMethod("""
        select id,category,account,account_type as accountType,balance ,priority
        from account 
        where category='credit' and id in (select account_id from strategy where enable=1 and shares_code=':sharesCode')""",
                                          Account)
    updateBalance = UpdateMethod("update account set balance = ':balance' where id=:id")

    addBalance = UpdateMethod("update account set balance = balance + :balance where id=:id")

    addBalanceByStrategyId = UpdateMethod(
        "update account set balance = balance + :balance where id=(select account_id from strategy where id=:strategyId)  and exists (select id from trade_log where id=:id and status in (:status))")

    updateBalanceByTradeLog = UpdateMethod("""
        update account a,(select b.account_id,sum(a.order_count*a.price) amount from trade_log a 
                inner join strategy b on a.strategy_id=b.id where b.shares_code=':sharesCode' and a.`type`='BUY' and (status=0 or status is null or order_no is null) group by b.account_id) b
        set a.balance=a.balance+b.amount   
        where a.id=b.account_id     
    """)


class AccountPositionDao(BaseDao):
    __mode__ = AccountPosition
    # deleteAll = UpdateMethod("delete from account_position",
    #                          "DELETE FROM sqlite_sequence WHERE `name` = 'account_position'")
    deleteAll = UpdateMethod("truncate account_position")

    replaceSave = UpdateMethod("""
        replace into account_position(account_id,shares_code,enable_count)
        values(':accountId',':sharesCode',':enableCount')
    """)

    querySharesCodeByAccountId = QueryMethod(
        "select distinct  shares_code from account_position where account_id=:accountId", str)

    selectOneByAccountIdAndSharesCode = SelectMethod(
        "select id,account_id as accountId,shares_code as sharesCode,enable_count as enableCount from account_position where account_id=:accountId and shares_code=':sharesCode'",
        AccountPosition)

    updateEnableCountByAccountIdAndSharesCode = UpdateMethod(
        "update account_position set enable_count = enable_count+:num where account_id=:accountId and shares_code=':sharesCode'")

    updateEnableCountByAccountIdAndSharesCodeAndLogIdAndStatus = UpdateMethod(
        "update account_position set enable_count = enable_count+:num where account_id=:accountId and shares_code=':sharesCode' and exists (select id from trade_log where id=:id and status in (:status))")

    updateEnableCountByLogId = UpdateMethod("""
        update account_position a ,trade_log b , strategy c set enable_count = enable_count + b.order_count where a.shares_code=c.shares_code and a.account_id=c.account_id and c.id=b.strategy_id and b.`type`='BUY' and b.id = :logId
    """)

    updateEnableCountByTradeLog = UpdateMethod("""
        update account_position a ,(select b.account_id,b.shares_code,sum(a.order_count) `count` from trade_log a 
        inner join strategy b on a.strategy_id=b.id where b.shares_code=':sharesCode' and a.`type`='SELL' and (status=0 or status is null or order_no is null) group by b.account_id,b.shares_code) b
        set a.enable_count=a.enable_count+b.`count` 
        where a.account_id=b.account_id and a.shares_code=b.shares_code
    """)

    queryAssistData = SelectMethod("""
        select b.shares_code as sharesCode,a2.balance ,ap.enable_count as enableCount,
            ifnull(sum(case when a.`type`='BUY' and a.status in (0,2)  then a.order_count *a.price else 0 end),0) buyAmount,
            ifnull(sum(case when a.`type`='BUY' and a.status in (2) then a.order_count else 0 end),0) buyCount,
            ifnull(sum(case when a.`type` = 'SELL' and a.status in (2)  then  a.order_count*a.price end),0) sellAmount,
            ifnull(sum(case when a.`type`='SELL' and a.status in (0,2) then a.order_count else 0 end),0) sellCount
        from trade_log a
        inner join strategy b on a.strategy_id=b.id 
        inner join account_position ap on ap.shares_code = b.shares_code 
        inner join account a2 on a2.id = ap.account_id and a2.id = b.account_id 
        where b.shares_code=':sharesCode'
        group by b.shares_code,a2.balance ,ap.enable_count
    """, AssistDto)


class RevokeOrderLogDao(BaseDao):
    __mode__ = RevokeOrderLog
    selectOneByOrderNo = SelectMethod(
        "select id,order_no as orderNo,`time`,`status` from revoke_order_log where order_no=':orderNo' and `status` in (0,1) order by `time` desc limit 1",
        RevokeOrderLog)
    queryByOrderNo = QueryMethod(
        "select order_no from revoke_order_log where order_no in (:orderNos) and `status` in (0,1)", str)
    cleanHistoryData = UpdateMethod(
        "delete from revoke_order_log where date_format(now(),'%Y-%m-%d')>date_format(`time`,'%Y-%m-%d')")
    updateStatusByOrderNo = UpdateMethod("update revoke_order_log set `status`=:status where order_no=':orderNo'")

    deleteByOrderNo = UpdateMethod("delete from revoke_order_log where order_no = ':orderNo'")

    deleteByTodayBefore = UpdateMethod(
        "delete from revoke_order_log where date_format(now(),'%Y-%m-%d') > date_format(`time`,'%Y-%m-%d')")


class SharesQuotationDao(BaseDao):
    __mode__ = SharesQuotation

    deleteByTodayBefore = UpdateMethod(
        "delete from shares_quotation where date_format(now(),'%Y-%m-%d') > date_format(create_time,'%Y-%m-%d')")

    deleteBySharesCode = UpdateMethod("delete from shares_quotation where shares_code=':sharesCode'")

    cleanAll = UpdateMethod("delete from shares_quotation")


class SharesQuotationHistoryDao(BaseDao):
    __mode__ = SharesQuotationHistory

    insertSelect = UpdateMethod("""
        insert into shares_quotation_history(id,shares_code,last_price,high_price,open_price,pre_close_price,time,create_time) 
        select id,shares_code,last_price,high_price,open_price,pre_close_price,time,create_time 
        from shares_quotation where date_format(now(),'%Y-%m-%d') > date_format(create_time,'%Y-%m-%d')""")

    selectCount = SelectMethod(
        "select count(1) from shares_quotation_history where shares_code in (:sharesCodes) and last_price != 0 and date_format(create_time,'%Y-%m-%d') > ':startDate' and date_format(create_time,'%Y-%m-%d') <= ':endDate'", int)

    queryByLastpriceNotNone = QueryMethod("""
        select 0 as stopMark,shares_code as symbol,high_price as highPrice,last_price as lastPrice,
        date_format(create_time,'%Y-%m-%d') as time,open_price as openPrice,pre_close_price as preClosePrice
        from shares_quotation_history 
        where shares_code in (:sharesCodes) and last_price != 0 and open_price != 0 and pre_close_price != 0 and date_format(create_time,'%Y-%m-%d') > ':startDate' and date_format(create_time,'%Y-%m-%d') <= ':endDate'
        order by id limit :limit offset :offset""", RealMarketData)


class StrategyDao(BaseDao):
    __mode__ = Strategy

    enableByIds = UpdateMethod("update strategy set enable=1, stop_time=null where id in (:ids)")

    queryIdByKey = QueryMethod("""
        select max(id) from strategy where concat(account_id,',',shares_code,',',version) in (:keys)
    """, int)

    deleteBySharesCode = UpdateMethod("delete from strategy where shares_code=':sharesCode'")

    queryCountByEnableSell = QueryMethod("""
        select shares_code as sharesCode, version,sum(case when enable_sell=1 then 1 else 0 end) as startCount,sum(case when enable_sell=0 then 1 else 0 end) as stopCount
        from strategy 
        where shares_code like ':sharesCode'
        group by shares_code, version
        order by shares_code
    """, SellStatusDto)

    disable = UpdateMethod(
        "update strategy set enable=0,stop_time=date_format(now(),'%Y-%m-%d %H:%i:%S') where version=':version' and `name`=':name'")
    queryStrategyByNotSellAll = QueryMethod("""
        select a.id,a.`name`,a.account_id as accountId,a.shares_code as sharesCode,a.create_time as createTime,a.json,a.version,a.enable,a.enable_sell as enableSell
        from strategy a
        left outer join (select strategy_id,sum(`buy_count`) as ct from trade_log where `type` in ('BUY','MARGIN_BUY','DEBIT_BUY') and status in(1,2,3) group by strategy_id)  b on a.id=b.strategy_id
        left outer join (select strategy_id,sum(`buy_count`) as ct from trade_log where `type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and status in(0,1,2,3) group by strategy_id)  c on a.id=c.strategy_id
        where ((b.ct-c.ct)>0 and a.enable=0) or enable=1 or (a.version='1.4' and a.enable_sell = 1)""", Strategy)

    disableByAccountIdAndSharesCodeAndVersion = UpdateMethod("""
        update strategy set enable=0,stop_time=date_format(now(),'%Y-%m-%d %H:%i:%S') 
        where concat(account_id,',',shares_code,',',version) in (:keys) and enable=1""")

    disableBySharesCode = UpdateMethod("""
        update strategy set enable=0,stop_time=date_format(now(),'%Y-%m-%d %H:%i:%S') 
        where shares_code = ':sharesCode' and enable=1 and version= ':version'
    """)

    queryBySharesCode = QueryMethod("""
        select a.id,a.`name`,a.account_id as accountId,a.shares_code as sharesCode,a.create_time as createTime,a.json,a.version,a.enable,a.enable_sell as enableSell,t.minPrice
        from strategy a
        left join (select strategy_id,min(price) as minPrice from trade_log 
                    where id not in (select buy_id from trade_log where `type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and status in (0,1,2,3) and buy_id is not null)
                        and `type` in ('BUY','MARGIN_BUY','DEBIT_BUY') and status in(1,2,3) group by strategy_id) t on a.id=t.strategy_id
        left outer join (select strategy_id,sum(`buy_count`) as ct from trade_log where `type` in ('BUY','MARGIN_BUY','DEBIT_BUY') and status in(1,2,3) group by strategy_id)  b on a.id=b.strategy_id
        left outer join (select strategy_id,sum(`buy_count`) as ct from trade_log where `type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and status in(0,1,2,3) group by strategy_id)  c on a.id=c.strategy_id
        where shares_code = ':sharesCode' and (((b.ct-ifnull(c.ct,0))>0 and a.enable=0) or enable=1 or (a.version='1.4' and a.enable_sell=1)) order by t.minPrice""",
                                    Strategy)

    queryBySharesCodeAndAccountId = QueryMethod("""
            select id,`name`,account_id as accountId,shares_code as sharesCode,create_time as createTime,json,version,enable,enable_sell as enableSell 
            from strategy
            where account_id=':accountId' and shares_code like ':sharesCode' and enable=1 and version=':version'
            union 
            select a.id,a.`name`,a.account_id as accountId,a.shares_code as sharesCode,a.create_time as createTime,a.json,a.version,a.enable,a.enable_sell as enableSell 
            from strategy a 
            inner join (select shares_code,max(id) id from strategy where enable=0  and version=':version' group by shares_code) s on a.shares_code = s.shares_code and a.id=s.id
            where a.account_id=':accountId' and a.shares_code like ':sharesCode' and a.shares_code  not in (select shares_code from strategy where enable=1  and version=':version')
    """, Strategy)

    getById = SelectMethod(
        "select id,`name`,account_id as accountId,shares_code as sharesCode,create_time as createTime,json,version,enable,enable_sell as enableSell from strategy where id=:id",
        Strategy)

    querySharesCodeByAccountId = QueryMethod("""
        select distinct shares_code
        from strategy
        where account_id=:accountId""", str)

    updateEnableSellBySharesCode = UpdateMethod(
        "update strategy set enable_sell=':status' where shares_code in (:sharesCodes) and version=':version'")

    updateEnableSellById = UpdateMethod(
        "update strategy set enable_sell=':status' where id=:strategyId")

    updateJsonBySharesCode = UpdateMethod("update strategy set json=':json' where shares_code=':sharesCode'")

    updateJsonById = UpdateMethod("update strategy set json=':json' where id=':id'")

    selectBySharesCodeAndEnable = QueryMethod("""
        select a.id,a.`name`,a.account_id as accountId,a.shares_code as sharesCode,a.create_time as createTime,a.json,a.version,a.enable,a.enable_sell as enableSell
        from strategy a 
        where a.shares_code=':sharesCode' and a.enable=1
    """, Strategy)

    countTodayUpdate = SelectMethod(
        "select count(1) from strategy where shares_code=':sharesCode' and version='1.2' and date_format(create_time,'%Y-%m-%d')=date_format(now(),'%Y-%m-%d')",
        int)

    selectBySharesCodeAndAccountIdAndEnable = SelectMethod("""
        select a.id,a.`name`,a.account_id as accountId,a.shares_code as sharesCode,a.create_time as createTime,a.json,a.version,a.enable,a.enable_sell as enableSell
        from strategy a 
        where a.shares_code=':sharesCode' and a.enable=:enable and a.account_id=:accountId and a.version=':version'
        order by a.id desc
        limit 1
    """, Strategy)


class OpenTradeLogDao(BaseDao):
    __mode__ = OpenTradeLog

    selectByStrategyIdAndDate = SelectMethod(
        "select id,strategy_id as strategyId,date,log_id as logId from open_trade_log where strategy_id=':strategyId' and date=':date'",
        OpenTradeLog)

    cleanAll = UpdateMethod("delete from open_trade_log")


class WorkerRegisterDao(BaseDao):
    __mode__ = WorkerRegister

    countWorker = SelectMethod(
        "select count(1) from worker_register where date_add(heart_time, interval 10 second) > now()", int)

    deleteExpire = UpdateMethod("delete from worker_register where date_add(heart_time, interval 10 second) < now()")

    updateHeartTime = UpdateMethod(
        "update worker_register set heart_time=date_format(now(),'%Y-%m-%d %H:%i:%S') where worker_key=':key'")


class TradeLogDao(BaseDao):
    __mode__ = TradeLog

    sumSellCountByStrategyId = SelectMethod("select ifnull(sum(buy_count),0) from trade_log where strategy_id=':strategyId' and `type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and status in (2,3)", int)

    selectLatestByStrategyId = SelectMethod("""select id,strategy_id as strategyId,`type`,status,price,power,avg_price as avgPrice,`order_count` as orderCount,`buy_count` as buyCount,`time`,order_no as orderNo,buy_id as buyId,trade_time as tradeTime
        from trade_log
        where strategy_id=':strategyId' and `type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and status in (2,3)
        order by id desc 
        limit 1
        """, TradeLog)

    selectSellCountByStrategyId = SelectMethod("""select ifnull(sum(buy_count), 0) from trade_log where strategy_id=':strategyId' and status in (2, 3) and `type` in ('SELL','DEBIT_SELL','MARGIN_SELL')""", int)

    selectOneByStrategyIdAndStatus = SelectMethod("""select id,strategy_id as strategyId,`type`,status,price,power,avg_price as avgPrice,`order_count` as orderCount,`buy_count` as buyCount,`time`,order_no as orderNo,buy_id as buyId,trade_time as tradeTime
        from trade_log
        where strategy_id=':strategyId' and `type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and date_format(`time`,'%Y-%m-%d')=date_format(now(),'%Y-%m-%d') and status in (0, 1)
        limit 1
        """, TradeLog)

    selectByStrategyId = SelectMethod("""select id,strategy_id as strategyId,`type`,status,price,power,avg_price as avgPrice,`order_count` as orderCount,`buy_count` as buyCount,`time`,order_no as orderNo,buy_id as buyId,trade_time as tradeTime
        from trade_log
        where strategy_id=':strategyId' and `type` in ('BUY','MARGIN_BUY','DEBIT_BUY') and date_format(`time`,'%Y-%m-%d')=date_format(now(),'%Y-%m-%d') and status in (0, 1)
        order by id desc 
        limit 1
        """, TradeLog)

    deleteBySharesCode = UpdateMethod("delete from trade_log where strategy_id in (select id from strategy where shares_code = ':sharesCode')")

    cleanAll = UpdateMethod("delete from trade_log")

    cleanData = UpdateMethod(
        "delete from trade_log where date_format(now(),'%Y-%m-%d') > date_format(`time`,'%Y-%m-%d') and (status=0 or status is null or order_no is null)")

    cleanDataRegression = UpdateMethod(
        "delete from trade_log where  (status=0 or status is null or order_no is null) and strategy_id in (select id from strategy where shares_code = ':sharesCode')")

    queryTodayBuyLogByStrategyId = QueryMethod("""select id,strategy_id as strategyId,`type`,status,price,power,avg_price as avgPrice,`order_count` as orderCount,`buy_count` as buyCount,`time`,order_no as orderNo,buy_id as buyId,trade_time as tradeTime
        from trade_log
        where strategy_id =:id  and `type` in ('BUY','MARGIN_BUY','DEBIT_BUY') and date_format(`time`,'%Y-%m-%d')=date_format(now(),'%Y-%m-%d')
        order by `time` desc""", TradeLog)

    updateOrderNoAndStatusById = UpdateMethod(
        "update trade_log set `status`=:status where id=:id and `status` < :status",
        "update trade_log set order_no=':orderNo' where id=:id")

    getById = SelectMethod("""select id,strategy_id as strategyId,`type`,status,price,power,avg_price as avgPrice,`order_count` as orderCount,`buy_count` as buyCount,`time`,order_no as orderNo,buy_id as buyId,trade_time as tradeTime
        from trade_log
        where id=:id""", TradeLog)

    updateStatusAndBuyCountByOrderNo = UpdateMethod(
        "update trade_log set `status`=:status,buy_count=:buyCount,avg_price=:avgPrice,trade_time=':tradeTime',order_no=':orderNo' where id=:id")

    getByOrderNoAndStatus = SelectMethod("""select id,strategy_id as strategyId,`type`,status,price,power,avg_price as avgPrice,`order_count` as orderCount,`buy_count` as buyCount,`time`,order_no as orderNo,buy_id as buyId,trade_time as tradeTime
        from trade_log
        where order_no=':orderNo'""", TradeLog)

    deleteById = UpdateMethod("delete from trade_log where id=:id")

    queryByStatus = QueryMethod("""select c.account as account,c.account_type as accountType,a.id,a.strategy_id as strategyId,a.`type`,a.status,a.price,a.power,a.avg_price as avgPrice,a.`order_count` as orderCount,a.`buy_count` as buyCount,a.`time`,a.order_no as orderNo,a.buy_id as buyId,a.trade_time as tradeTime
        from trade_log a
        inner join strategy b on a.strategy_id=b.id
        inner join account c on b.account_id=c.id
        where a.status in (:status) and a.order_no is not null """, AccountTradeLogDto)

    selectOneByStrategyIdAndClosestPriceLog = QueryMethod("""select * from (
          select a.id,a.strategy_id as strategyId,a.`type`,a.status,a.price,a.power,a.avg_price as avgPrice,a.`order_count` as orderCount,a.`buy_count` as buyCount,a.`time`,a.order_no as orderNo,a.buy_id as buyId,
            (select ifnull(max(last_price),:highPrice)-:incA from shares_quotation where shares_code=b.shares_code and date_format(`create_time`,'%Y-%m-%d %H-%i-%S') >= date_format(a.`trade_time`,'%Y-%m-%d %H-%i-%S')) as highPrice
            from trade_log a
            inner join strategy b on a.strategy_id=b.id
            where a.strategy_id=:strategyId and a.`type` in ('BUY','MARGIN_BUY','DEBIT_BUY') and a.status in (2,3)
        ) c where :highPrice - :incA>=c.price and c.highPrice>=c.price
		order by c.price desc""", TradeLog)

    queryMarketableCountBySharesCode = SelectMethod("""select sum(buyCount) as `count`
        from (
        select ifnull(sum(a.buy_count),0) buyCount
        from trade_log a
        inner join strategy b on a.strategy_id=b.id
        where date_format(now(),'%Y-%m-%d')>date_format(a.`time`,'%Y-%m-%d') and a.`type` in ('BUY','MARGIN_BUY','DEBIT_BUY') and a.status in(1,2,3) and b.shares_code=':sharesCode'
        union all
        select ifnull(-sum(a.buy_count),0) buyCount
        from trade_log a
        inner join strategy b on a.strategy_id=b.id
        where a.`type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and a.status in(0,1,2,3) and b.shares_code=':sharesCode'
        ) a""", int)

    queryNotSellOrderByStrategyIdAndLtCurPrice = QueryMethod("""select a.id,a.strategy_id as strategyId,a.`type`,a.status,a.price,a.power,a.avg_price as avgPrice,a.`order_count` as orderCount,a.`buy_count`-ifnull(b.buy_count,0) as buyCount,a.`time`,a.order_no as orderNo,a.buy_id as buyId
        from trade_log a
        left OUTER join (select buy_id,sum(case when status in (0,1) then order_count when status in (2,3) then buy_count end) as buy_count from trade_log where buy_id is not null and `type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and status in(0,1,2,3) group by buy_id) b on a.id=b.buy_id
        where a.`type` in ('BUY','MARGIN_BUY','DEBIT_BUY') and a.status in(2,3) and a.strategy_id=':strategyId' and :curPrice>=a.avg_price and a.`buy_count`-ifnull(b.buy_count,0)>0
        order by a.avg_price""", TradeLog)

    deleteByBuyId = UpdateMethod("delete from trade_log where buy_id=:buyId")

    queryNotSellByStrategyIdAndGtCurPrice = QueryMethod("""select a.id,a.strategy_id as strategyId,a.`type`,a.status,a.price,a.power,a.avg_price as avgPrice,a.`order_count` as orderCount,a.`buy_count` as buyCount,a.`time`,a.order_no as orderNo,a.buy_id as buyId,a.trade_time as tradeTime
        from trade_log a
        where a.strategy_id in (select id from strategy where shares_code=':sharesCode' and version=':version') and  a.price >:curPrice and a.`type` in ('SELL','DEBIT_SELL','MARGIN_SELL') and a.status = 0 and a.order_no is not null and a.order_no not in (select order_no from revoke_order_log)
        order by a.price desc""", TradeLog)

    updateBuyIdById = UpdateMethod("update trade_log set buy_id = :buyId where id=:id")

    updateBuyIdIsNullByBuyId = UpdateMethod("update trade_log set buy_id = null where buy_id=:buyId")

    querySpTimes = QueryMethod("""select distinct date(`time`) as times
        from trade_log
        where date_format(now(),'%Y-%m-%d') > date_format(`time`,'%Y-%m-%d') and strategy_id=:strategyId
        order by 1 desc
        limit :spTimes""", str)

    queryCountByPriceAndDate = SelectMethod("""select count(1) as ct
        from trade_log
        where date_format(`time`,'%Y-%m-%d') in (:dates ) and price=:price and `status` in (2,3) and strategy_id=:strategyId""",
                                            int)
    statisticsRate = QueryMethod("""
        select b.shares_code as sharesCode,sum(c.buy_count*c.avg_price) buyAmount,sum(a.buy_count*a.avg_price) sellAmount
        from trade_log a 
        inner join strategy b on a.strategy_id=b.id 
        inner join trade_log c on a.buy_id=c.id and c.strategy_id = b.id 
        where b.shares_code in (:sharesCodes) and a.status = 2 and a.`type` = 'SELL'
        group by b.shares_code
    """, YieldDataDto)


class HistoryDataDao(BaseDao):
    __mode__ = HistoryData


class TradeLogTestDao(BaseDao):
    __mode__ = TradeLogTest

    insertSelectByTradeLog = UpdateMethod("""
        insert into trade_log_test(id,strategy_id,type,status,price,power,avg_price,order_count,time,trade_time,order_no,buy_count,buy_id,group_id)
        select id,strategy_id,type,status,price,power,avg_price,order_count,time,trade_time,order_no,buy_count,buy_id,:groupId 
        from trade_log 
        """)


class YieldLogDao(BaseDao):
    __mode__ = YieldLog

    selectMaxGroupId = SelectMethod("select ifnull(max(group_id),0) from yield_log", int)


class AccountAssetsLogDao(BaseDao):
    __mode__ = AccountAssetsLog

    selectCount = SelectMethod("select count(1) from account_assets_log", int)

    selectNewest = SelectMethod("""
        select id,account_id as accountId, `date` , assets, shares, net_value as netValue, pre_net_value as preNetValue
        from account_assets_log
        order by `date` desc 
        limit 1
    """, AccountAssetsLog)

    queryPageLog = QueryMethod("""
        select id,account_id as accountId, `date` , assets, shares, net_value as netValue, pre_net_value as preNetValue
        from account_assets_log
        order by `date` desc 
        limit :limit offset :offset
    """, AccountAssetsLog)

    queryLogByGtDate = QueryMethod("""
        select id,account_id as accountId, `date` , assets, shares, net_value as netValue, pre_net_value as preNetValue
        from account_assets_log
        where `date`>=':date' and `date` <= ':endDate'
        group by `date`
        order by `date`
        """, AccountAssetsLog)

    queryLogByBtDate = QueryMethod("""
        select `date`,assets, shares, net_value as netValue,pre_net_value as preNetValue
        from account_assets_log
        where `date` >= ':qStart' and `date` <= ':qEnd' 
        order by `date` desc
        """, AccountAssetsLog)

    sumShares = SelectMethod(
        "select ifnull(sum(shares),0) from account_assets_log where `date`<':date'", float)

    selectByDate = SelectMethod("""
        select id,account_id as accountId, `date` , assets, shares, net_value as netValue, pre_net_value as preNetValue
        from account_assets_log
        where `date` = ':date'
        """, AccountAssetsLog)

    selectByEtDate = SelectMethod("""
        select id,account_id as accountId, `date` , assets, shares, net_value as netValue, pre_net_value as preNetValue
        from account_assets_log
        where `date` < ':etDate' 
        order by `date` desc
        limit 1
        """, AccountAssetsLog)

    updateSharesByDate = UpdateMethod(
        "update account_assets_log set shares=:shares where `date`=':date' and account_id=:accountId")

    clearSharesByDate = UpdateMethod(
        "update account_assets_log set shares=0 where `date` in (:dates) and account_id=:accountId")

    updateAssetsAndNetValueById = UpdateMethod("update account_assets_log set assets=:assets, net_value=:netValue where id=:id")

class ContrastLineDataDao(BaseDao):
    __mode__ = ContrastLineData
    queryByDate = QueryMethod("""
        select `date`, shares_code as sharesCode , `close`, `pre_close` as preClose
        from contrast_line_data
        where `date` >= ':date'
        order by `date`
    """, ContrastLineData)

    deleteByDate = UpdateMethod(
        "delete from contrast_line_data where `date` >=':begin' and `date`<=':end' and shares_code = ':sharesCode'")

    selectMaxDate = SelectMethod("select max(`date`) from contrast_line_data where `date` >= ':date'", str)


class LoginUserDao(BaseDao):
    __mode__ = LoginUser

    selectByLoginName = SelectMethod("""
        select id, login_name as loginName, pwd, user_name as userName, enable, create_time as createTime, login_count as loginCount, update_pwd as updatePwd
        from login_user
        where login_name = ':loginName'
        """, LoginUser)

    addLoginCountByLoginName = UpdateMethod(
        "update login_user set login_count=login_count + 1 where login_name=':loginName'")

    queryByUserName = QueryMethod("""
        select id, login_name as loginName, pwd, user_name as userName, enable, create_time as createTime, login_count as loginCount, update_pwd as updatePwd
        from login_user
        where user_name like ':userName'
        """, LoginUser)

    updateEnableByLoginName = UpdateMethod("update login_user set enable=:enable where login_name=':loginName'")

    updatePwdByLoginName = UpdateMethod("update login_user set pwd=':newPwd' where login_name=':loginName'")

    updateUpdatePwdByLoginName = UpdateMethod(
        "update login_user set update_pwd=:updatePwd where login_name=':loginName'")

    queryByIds = QueryMethod("""
        select id, login_name as loginName, pwd, user_name as userName, enable, create_time as createTime, login_count as loginCount, update_pwd as updatePwd
        from login_user
        where id in (:ids)
    """, LoginUser)


class UserInvestmentLogDao(BaseDao):
    __mode__ = UserInvestmentLog

    queryLogByBtDate = QueryMethod("""
        select a.id, a.time, b.user_name as userName, TRUNCATE(a.transfer_amount,2) as amount, truncate(a.shares,2) as shares, truncate(a.buy_net_value,4) as netValue
        from user_investment_log a
        inner join login_user b on a.user_id = b.id
        where date_format(a.`time`, '%Y-%m-%d') >= ':qStart' and date_format(a.`time`, '%Y-%m-%d') <= ':qEnd' and a.account_id = :accountId and b.user_name like ':queryKey'
        order by a.`time` desc 
    """, UserInvestmentDto)

    deleteByIds = UpdateMethod("delete from user_investment_log where id in (:ids)")

    sumUserAmountByEtDate = SelectMethod("""
        select TRUNCATE(ifnull(sum(transfer_amount),0),2) from user_investment_log 
        where date_format(`time`, '%Y-%m-%d %H:%i:%S') <=':etDate' and assets_log_id is null
    """, float)

    queryLogByEtDate = QueryMethod("""
        select id, account_id as accountId, user_id as userId, assets_log_id as assetsLogId, TRUNCATE(transfer_amount,2) as transferAmount, shares, buy_net_value as buyNetValue, `time`
        from user_investment_log
        where date_format(`time`, '%Y-%m-%d %H:%i:%S') <=':etDate' and assets_log_id is null
    """, UserInvestmentLog)

    updateSharesAndNetValueByEtDate = UpdateMethod("""
        update user_investment_log set shares=:shares, buy_net_value=:netValue, assets_log_id=:assetsLogId
        where date_format(`time`, '%Y-%m-%d %H:%i:%S') <= ':etDate' and assets_log_id is null
    """)

    queryLogByUserId = QueryMethod("""
        select id, account_id as accountId, user_id as userId, assets_log_id as assetsLogId, TRUNCATE(transfer_amount,2) as transferAmount, shares, buy_net_value as buyNetValue, `time`
        from user_investment_log
        where user_id=:userId
        order by `time` desc 
    """, UserInvestmentLog)

    queryPageLogByUserId = QueryMethod("""
        select id, account_id as accountId, user_id as userId, assets_log_id as assetsLogId, TRUNCATE(transfer_amount,2) as transferAmount, shares, buy_net_value as buyNetValue, `time`
        from user_investment_log
        where user_id=:userId
        order by `time` desc 
        limit :limit offset :offset
    """, UserInvestmentLog)

    selectCountByUserId = SelectMethod("select count(1) from user_investment_log where user_id=:userId", int)

@transaction
def test():
    accounts = [AccountPosition(accountId=1, sharesCode='123444', enableCount=1),
                AccountPosition(accountId=1, sharesCode='123555', enableCount=1)]
    accPositionDao = AccountPositionDao()
    accPositionDao.create()
    accPositionDao.insert(accounts)
    scStrs = accPositionDao.querySharesCodeByAccountId({"accountId": 1})
    ac = accPositionDao.selectOneByAccountIdAndSharesCode({"accountId": 1, "sharesCode": "123555"})
    accPositionDao.updateEnableCountByAccountIdAndSharesCode({"num": 1, "accountId": 1, "sharesCode": "123555"})
    # accPositionDao.deleteAll()

    print(scStrs)


if __name__ == '__main__':
    test()




from kindenty.base.BaseMethod import QueryMethod, UpdateMethod, SelectMethod
from kindenty.base.DatabaseDriver import driver

from kindenty.base.ExceptionModel import ArgsException
from kindenty.base.OrmField import IDField



class DaoMetaclass(type):

    def __new__(cls, name, bases, attrs):
        if cls.__name__ == 'BaseDao':
            return type.__new__(cls, name, bases, attrs)
        mappings = dict()
        for k, v in attrs.items():
            if isinstance(v, (QueryMethod, UpdateMethod, SelectMethod)):
                mappings[k] = v
        for k, v in mappings.items():
            attrs.pop(k)
            if isinstance(v, (QueryMethod, SelectMethod)):
                attrs[k] = lambda self, param={}, method=v: driver.result(driver.execute([method.sql], param), method)
            elif isinstance(v, UpdateMethod):
                attrs[k] = lambda self, param={}, sql=v.sql: driver.update(driver.execute([*sql], param))
        attrs['insert'] = lambda self, models=list: driver.update(
            driver.insert(self, driver.execute(*self.__insert__(models)), *models))
        attrs['create'] = lambda self: driver.execute(*self.__create__())
        attrs['queryAll'] = lambda self: driver.result(driver.execute(*self.__query_all__()),
                                                       QueryMethod('', self.__mode__))
        attrs['__mappings__'] = mappings

        return type.__new__(cls, name, bases, attrs)


class BaseDao(dict, metaclass=DaoMetaclass):

    def __create__(self):
        fields = []
        for k, v in self.__mode__.__mappings__.items():
            fields.append('%s %s' % (v.name, v.columnType))
        sql = 'CREATE TABLE IF NOT EXISTS %s (%s)' % (self.__mode__.__table__, ','.join(fields))
        return [sql], []

    def __insert__(self, models: list):
        if models is None or len(models) == 0:
            raise ArgsException('%s model defined error' % self.name)
        fields = []
        params = []
        args = []
        prop = []
        for k, v in self.__mode__.__mappings__.items():
            if isinstance(v, IDField):
                continue
            fields.append(v.name)
            params.append('?')
            prop.append(k)
        values = []
        for m in models:
            values.append('(%s)' % ','.join(params))
            for k, v in self.__mode__.__mappings__.items():
                if isinstance(v, IDField):
                    continue
                args.append(v.parseToValue(getattr(m, k)))
        sql = 'insert into %s (%s) values %s' % (self.__mode__.__table__, ','.join(fields), ','.join(values))
        return [sql], args

    def __query_all__(self):
        fields = []
        for k, v in self.__mode__.__mappings__.items():
            fields.append('%s as %s' % (v.name, k))
        sql = 'select %s from %s' % (','.join(fields), self.__mode__.__table__)
        return [sql], []


def transaction(fun):
    return driver.transaction(fun)



from enum import Enum, unique

@unique
class DataType(Enum):
    int = 'INTEGER'
    text = 'TEXT'

    sqlite_str = 'CHARACTER'
    sqlite_float = 'DOUBLE'
    sqlite_date = 'CHARACTER(30)'

    mysql_str = 'VARCHAR'
    mysql_float = 'DECIMAL(18,6)'
    mysql_date = 'VARCHAR(30)'
    mysql_bigint = 'bigint'


class FieldType(object):
    __slots__ = ('str', 'int', 'float', 'boolean', 'datetime', 'bigStr','bigint')

    def __init__(self, **kw):
        self.str = kw.get('str', DataType.sqlite_str.value)
        self.int = kw.get('int', DataType.int.value)
        self.float = kw.get('float', DataType.sqlite_float.value)
        self.boolean = kw.get('boolean', DataType.int.value)
        self.datetime = kw.get('datetime', DataType.sqlite_date.value)
        self.bigStr = kw.get('bigStr', DataType.text.value)
        self.bigint = kw.get('bigint', DataType.int.value)


@unique
class DatabaseType(Enum):
    sqlite = FieldType()
    mysql = FieldType(str=DataType.mysql_str.value, int=DataType.int.value, float=DataType.mysql_float.value,
                      boolean=DataType.int.value, datetime=DataType.mysql_date.value, bigStr=DataType.text.value, bigint=DataType.mysql_bigint.value)






import logging
from datetime import datetime
from logging.handlers import RotatingFileHandler
import os


def create_dir_not_exist(path):
    if not os.path.exists(path):
        os.mkdir(path)

create_dir_not_exist('logs')
rHandler = RotatingFileHandler("logs/%s-log.txt" % datetime.now().strftime('%Y%m%d'), maxBytes=500 * 1024 * 1024,
                               backupCount=100, encoding='UTF-8')
rHandler.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s: %(message)s")
rHandler.setFormatter(formatter)
console = logging.StreamHandler()
console.setLevel(logging.INFO)

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s: %(message)s',
                    handlers=[rHandler, console])

log = logging.getLogger()
log.addHandler(rHandler)
# formatter = logging.Formatter("%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s: %(message)s")
# # 第四步，将logger添加到handler里面
# rHandler = RotatingFileHandler("%s-log.txt" % datetime.now().strftime('%Y%m%d'), maxBytes=100 * 1024 * 1024, backupCount=10, encoding='UTF-8')
# rHandler.setLevel(logging.DEBUG)
# rHandler.setFormatter(formatter)
#
# console = logging.StreamHandler()
# console.setLevel(logging.INFO)
# console.setFormatter(formatter)
# log.addHandler(rHandler)
# log.addHandler(console)



from kindenty.base.BaseModel import BaseModel
from kindenty.base.OrmField import IDField, StringField, FloatField, IntField, DateTimeField, BooleanField, \
    DateTimeMilliField, BigStrField, BigIntField


class Account(BaseModel):
    __table__ = 'account'
    id = IDField('id')
    category = StringField('category')
    account = StringField('account')
    accountType = StringField('account_type')
    balance = FloatField('balance')
    priority = StringField('priority')


class AccountPosition(BaseModel):
    __table__ = 'account_position'
    id = IDField('id')
    accountId = IntField('account_id')
    sharesCode = StringField('shares_code')
    enableCount = IntField('enable_count')


class RevokeOrderLog(BaseModel):
    __table__ = 'revoke_order_log'
    id = IDField('id')
    orderNo = StringField('order_no')
    time = DateTimeField('time')
    status = BooleanField('status')


class SharesQuotation(BaseModel):
    __table__ = 'shares_quotation'
    id = IDField('id')
    sharesCode = StringField('shares_code')
    lastPrice = FloatField('last_price')
    highPrice = FloatField('high_price')
    openPrice = FloatField('open_price')
    preClosePrice = FloatField('pre_close_price')
    time = StringField('time')
    createTime = DateTimeMilliField('create_time')


class SharesQuotationHistory(BaseModel):
    __table__ = 'shares_quotation_history'
    id = IDField('id')
    sharesCode = StringField('shares_code')
    lastPrice = FloatField('last_price')
    highPrice = FloatField('high_price')
    openPrice = FloatField('open_price')
    preClosePrice = FloatField('pre_close_price')
    time = StringField('time')
    createTime = DateTimeMilliField('create_time')


class Strategy(BaseModel):
    __table__ = 'strategy'
    id = IDField('id')
    name = StringField('name')
    accountId = IntField('account_id')
    sharesCode = StringField('shares_code')
    createTime = DateTimeField('create_time')
    json = BigStrField('json')
    version = StringField('version')
    enable = BooleanField('enable')
    stopTime = DateTimeField('stop_time')
    enableSell = IntField('enable_sell')


class TradeLog(BaseModel):
    __table__ = 'trade_log'
    id = IDField('id')
    strategyId = IntField('strategy_id')
    type = StringField('type')
    status = IntField('status')
    price = FloatField('price')
    power = IntField('power')
    avgPrice = FloatField('avg_price')
    orderCount = IntField('order_count')
    time = DateTimeField('time')
    tradeTime = DateTimeField('trade_time')
    orderNo = StringField('order_no')
    buyCount = IntField('buy_count')
    buyId = IntField('buy_id')


class TradeLogTest(BaseModel):
    __table__ = 'trade_log_test'
    id = IDField('id')
    strategyId = IntField('strategy_id')
    type = StringField('type')
    status = IntField('status')
    price = FloatField('price')
    power = IntField('power')
    avgPrice = FloatField('avg_price')
    orderCount = IntField('order_count')
    time = DateTimeField('time')
    tradeTime = DateTimeField('trade_time')
    orderNo = StringField('order_no')
    buyCount = IntField('buy_count')
    buyId = IntField('buy_id')
    groupId = IntField('group_id')


class YieldLog(BaseModel):
    __table__ = 'yield_log'
    id = IDField('id')
    sharesCode = StringField('shares_code')
    rate = FloatField('rate')
    groupId = IntField('group_id')


class OpenTradeLog(BaseModel):
    __table__ = 'open_trade_log'
    id = IDField('id')
    strategyId = IntField('strategy_id')
    date = StringField('date')
    logId = IntField('log_id')


class WorkerRegister(BaseModel):
    __table__ = 'worker_register'
    id = IDField('id')
    workerKey = StringField('worker_key')
    heartTime = DateTimeField('heart_time')


class HistoryData(BaseModel):
    __table__ = 'history_data'
    id = IDField('id')
    sharesCode = StringField('shares_code')
    date = StringField('date')
    time = StringField('time')
    open = FloatField('open')
    high = FloatField('high')
    low = FloatField('low')
    close = FloatField('close')
    preClose = FloatField('pre_close')
    volume = BigIntField('volume')
    turnover = BigIntField('turnover')
    dayVolume = BigIntField('day_volume')


class AccountAssetsLog(BaseModel):
    __table__ = 'account_assets_log'
    id = IDField('id')
    accountId = IntField('account_id')
    date = StringField('date')
    assets = FloatField('assets')
    shares = FloatField('shares')
    netValue = FloatField('net_value')
    preNetValue = FloatField('pre_net_value')


class ContrastLineData(BaseModel):
    __table__ = 'contrast_line_data'
    id = IDField('id')
    sharesCode = StringField('shares_code')
    date = StringField('date')
    close = FloatField('close')
    preClose = FloatField('pre_close')


class LoginUser(BaseModel):
    __table__ = 'login_user'
    id = IDField('id')
    loginName = StringField('login_name')
    pwd = StringField('pwd')
    userName = StringField('user_name')
    enable = IntField('enable')
    loginCount = IntField('login_count')
    updatePwd = IntField('update_pwd')
    createTime = DateTimeField('create_time')


class UserInvestmentLog(BaseModel):
    __table__ = 'user_investment_log'
    id = IDField('id')
    accountId = IntField('account_id')
    userId = IntField('user_id')
    assetsLogId = IntField('assets_log_id')
    transferAmount = FloatField('transfer_amount')
    shares = FloatField('shares')
    buyNetValue = FloatField('buy_net_value')
    time = DateTimeField('time')



from typing import Any


class QueryMethod:
    __slots__ = ('sql', 'result')

    def __init__(self, sql: str, result: Any):
        self.sql = sql
        self.result = result


class SelectMethod:
    __slots__ = ('sql', 'result')

    def __init__(self, sql: str, result):
        self.sql = sql
        self.result = result


class UpdateMethod:
    __slots__ = ('sql', 'params')

    def __init__(self, *sql):
        self.sql = sql



import json
from ctypes import Array
from typing import Any

from kindenty.base.OrmField import Field


class DtoType:

    def __init__(self, objType):
        self.type = objType


class ObjType(DtoType):

    def __init__(self, objType):
        super(ObjType, self).__init__(objType)


class ListType(DtoType):

    def __init__(self, objType):
        super(ListType, self).__init__(objType)


class ModelMetaclass(type):

    def __new__(cls, name, bases, attrs):
        if cls.__name__ == 'BaseModel':
            return type.__new__(cls, name, bases, attrs)
        mappings = dict()
        for k, v in attrs.items():
            if isinstance(v, (Field, DtoType)):
                mappings[k] = v
        for k in mappings.keys():
            attrs.pop(k)

        attrs['__mappings__'] = mappings
        return type.__new__(cls, name, bases, attrs)


class BaseModel(dict, metaclass=ModelMetaclass):

    def __init__(self, **kwargs):
        super(BaseModel, self).__init__(**kwargs)

    def __getattr__(self, key) -> Any:
        return self.get(key, None)

    def __setattr__(self, name: str, value: Any) -> None:
        self[name] = value

    def setAttrs(self, keys: (list, tuple, Array), values: (list, tuple)):
        mappings = self.__mappings__
        for i in range(len(keys)):
            if keys[i] in mappings.keys():
                self[keys[i]] = mappings[keys[i]].parseToObj(values[i])
            else:
                self[keys[i]] = values[i]
        return self

    def decode(self, s):
        d = json.loads(s)
        obj = typeToObj(self.__class__, d)
        return obj

    def dict2Obj(self, d):
        for k, v in d.items():
            dtoT = self.__mappings__.get(k, None)
            if isinstance(dtoT, ObjType):
                if isinstance(v, str):
                    v = json.loads(v)
                obj = dtoT.type()
                setattr(self, k, obj.dict2Obj(v))
            elif isinstance(dtoT, ListType):
                if isinstance(v, str):
                    v = json.loads(v)
                obj = dtoT.type()
                setattr(self, k, list(map(obj.dict2Obj, v)))
            else:
                setattr(self, k, v)


def typeToObj(type,d):
    obj = type()
    for k, v in d.items():
        dtoT = type.__mappings__.get(k, None)
        if isinstance(dtoT, ObjType):
            if isinstance(v, str):
                v = json.loads(v)
            obj = dtoT.type()
            setattr(obj, k, obj.dict2Obj(v))
        elif isinstance(dtoT, ListType):
            if isinstance(v, str):
                v = json.loads(v)
            setattr(obj, k, list(map(lambda o: typeToObj(dtoT.type,o), v)))
        else:
            setattr(obj, k, v)
    return obj


from decimal import Decimal
from kindenty.databaseConfig import type

from kindenty.base.DatabaseType import DatabaseType
from datetime import datetime
import time


class Field(object):

    def __init__(self, name, column_type, objFun, valueFun=None):
        self.name = name
        self.columnType = column_type
        self.parseToObj = objFun
        if valueFun is None:
            valueFun = objFun
        self.parseToValue = valueFun

    def __str__(self):
        return '<%s:%s-%s>' % (self.__class__.__name__, self.name, self.columnType)


class IDField(Field):

    def __init__(self, name):
        super(IDField, self).__init__(name, '%s NOT NULL PRIMARY KEY %s ' % (
            type.value.int, 'AUTO_INCREMENT' if type == DatabaseType.mysql else 'AUTOINCREMENT'),
                                      lambda x: int(x))


class StringField(Field):

    def __init__(self, name, len: int = 100):
        super(StringField, self).__init__(name, '%s(%d)' % (type.value.str, len),
                                          lambda x: str(x) if x is not None else None)


class EnumField(Field):
    def __init__(self, name):
        super(EnumField, self).__init__(name, type.value.str, lambda x: str(x) if x is not None else None)


class IntField(Field):

    def __init__(self, name):
        super(IntField, self).__init__(name, type.value.int, lambda x: int(x) if x is not None else None)


class BigIntField(Field):

    def __init__(self, name):
        super(BigIntField, self).__init__(name, type.value.bigint, lambda x: int(x) if x is not None else None)


class FloatField(Field):

    def __init__(self, name):
        super(FloatField, self).__init__(name, type.value.float,
                                         lambda x: Decimal(str(x)).quantize(Decimal('0.000000')) if x is not None else None,
                                         lambda x: float(x.quantize(Decimal('0.000000'))) if x is not None else None)


class BooleanField(Field):

    def __init__(self, name):
        super(BooleanField, self).__init__(name, type.value.boolean, lambda x: int(x) if x is not None else None)


class DateTimeField(Field):

    def __init__(self, name):
        super(DateTimeField, self).__init__(name, type.value.datetime,
                                            lambda x: datetime.strptime(x,
                                                                        '%Y-%m-%d %H:%M:%S') if x is not None else None,
                                            lambda x: x.strftime('%Y-%m-%d %H:%M:%S') if x is not None else None)
class DateTimeMilliField(Field):

    def __init__(self, name):
        super(DateTimeMilliField, self).__init__(name, type.value.datetime,
                                            lambda x: datetime.strptime(x,
                                                                        '%Y-%m-%d %H:%M:%S.%f') if x is not None else None,
                                            lambda x: x.strftime('%Y-%m-%d %H:%M:%S.%f') if x is not None else None)


class BigStrField(Field):

    def __init__(self, name):
        super(BigStrField, self).__init__(name, type.value.bigStr, lambda x: str(x) if x is not None else None)


if __name__ == '__main__':
    # print(time.time().strftime('%Y-%m-%d %H:%M:%S'))
    print(time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time())))
    print(time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(None)))



from kindenty.base import log
from kindenty.base.BaseMethod import QueryMethod, SelectMethod
from kindenty.base.BaseModel import BaseModel
from kindenty.base.ExceptionModel import ModelException, ResultMultipleException
from kindenty.base.OrmField import IDField
from kindenty.databaseConfig import pool


class MysqlExecuteDriver:

    @staticmethod
    def insert(model, cursor, *args):
        try:
            mappings = model.__mode__.__mappings__
            idList = list(filter(lambda k: isinstance(mappings[k], IDField), mappings))
            if idList is None or len(idList) > 1:
                raise ModelException('%s model defined error' % model.name)
            idKey = idList[0]
            _id = cursor.lastrowid
            for m in args:
                setattr(m, idKey, _id)
                _id += 1
            return cursor
        except Exception as e:
            raise e

    @staticmethod
    def result(cursor, method):
        try:
            if not cursor:
                if isinstance(method, QueryMethod):
                    return []
                if isinstance(method, SelectMethod):
                    return None
            props = list(map(lambda p: p[0], cursor.description))
            rs = cursor.fetchall()

            if isinstance(method, QueryMethod):
                if len(rs) == 0:
                    return []
                if isinstance(method.result(), BaseModel):
                    return [method.result().setAttrs(props, x) for x in rs]
                else:
                    return [method.result(x[0]) for x in rs]
            if isinstance(method, SelectMethod):
                if len(rs) == 0:
                    return None
                if len(rs) > 1:
                    raise ResultMultipleException()
                if isinstance(method.result(), BaseModel):
                    return method.result().setAttrs(props, rs[0])
                else:
                    return method.result(rs[0][0])
        except Exception as e:
            raise e

    @staticmethod
    def execute(sql: list, *args):
        conn = pool.get()
        with conn.cursor() as cursor:
            log.debug('SQL: %s' % sql)
            log.debug('PARAMS: %s' % (str(args)))
            result = []
            try:
                for s in sql:
                    insert = False
                    if '?' in s:
                        s = s.replace('?', '%s')
                        insert = True
                    if len(args) == 1 and isinstance(args[0], dict):
                        keys = list(args[0].keys())
                        keys.sort(key=len, reverse=True)
                        for k in keys:
                            v = args[0][k]
                            if isinstance(v, (list, tuple)):
                                s = s.replace(':%s' % k, ','.join(map(lambda x: "'%s'" % x, v)))
                            else:
                                s = s.replace(':%s' % k, "%s" % v)
                    if insert:
                        cursor.execute(s, *args)
                    else:
                        cursor.execute(s)
                    result.append(cursor)
                return result[-1]
            except Exception as e:
                raise e

    @staticmethod
    def transaction(fun):
        def wrapper(*args, **kwargs):
            try:
                conn = pool.get()
                conn.autocommit(False)
                conn.begin()
                result = fun(*args, **kwargs)
                conn.commit()
                return result
            except Exception as e:
                conn.rollback()
                raise e

        return wrapper

    @staticmethod
    def update(cursor):
        log.debug('ROWS: %d' % cursor.rowcount)
        return cursor


class SQLite3Driver:

    @staticmethod
    def insert(model, cursor, *args):
        try:
            mappings = model.__mode__.__mappings__
            idList = list(filter(lambda k: isinstance(mappings[k], IDField), mappings))
            if idList is None or len(idList) > 1:
                raise ModelException('%s model defined error' % model.name)
            idKey = idList[0]
            _id = cursor.lastrowid
            for m in args[::-1]:
                setattr(m, idKey, _id)
                _id -= 1
            return cursor
        except Exception as e:
            raise e

    @staticmethod
    def result(cursor, method):
        try:
            props = list(map(lambda p: p[0], cursor.description))
            rs = cursor.fetchall()

            if isinstance(method, QueryMethod):
                if len(rs) == 0:
                    return []
                if isinstance(method.result(), BaseModel):
                    return [method.result().setAttrs(props, x) for x in rs]
                else:
                    return [method.result(x[0]) for x in rs]
            if isinstance(method, SelectMethod):
                if len(rs) == 0:
                    return None
                if len(rs) > 1:
                    raise ResultMultipleException()
                if isinstance(method.result(), BaseModel):
                    return method.result().setAttrs(props, rs[0])
                else:
                    return method.result(rs[0][0])
        except Exception as e:
            raise e

    @staticmethod
    def execute(sql: list, *args):
        conn = pool.get()

        log.debug('SQL: %s' % sql)
        log.debug('PARAMS: %s' % (str(args)))
        result = []
        try:
            for s in sql:
                if len(args) == 1 and isinstance(args[0], dict):
                    for k, v in args[0].items():
                        if isinstance(v, (list, tuple)):
                            s = s.replace(':%s' % k, ','.join(map(lambda x: "'%s'" % x, v)))
                result.append(conn.execute(s, *args))
            return result[-1]
        except Exception as e:
            raise e

    @staticmethod
    def transaction(fun):
        def wrapper(*args, **kwargs):
            try:
                conn = pool.get()
                conn.execute('BEGIN TRANSACTION')
                result = fun(*args, **kwargs)
                conn.commit()
                return result
            except Exception as e:
                conn.rollback()
                raise e

        return wrapper

    @staticmethod
    def update(cursor):
        log.debug('ROWS: %d' % cursor.rowcount)
        return cursor


driver = MysqlExecuteDriver



import setuptools

setuptools.setup(
    name="kindenty_base",
    version="0.1.4",
    author="kindenty",
    author_email="kindenty@163.com",
    description="测试",
    long_description="text/markdown",
    url="https://gitee.com/kindenty-yueyu/cats-strategy-core",
    packages=setuptools.find_packages(),
    classifiers=["Programming Language :: Python :: 3"],
)



from kindenty.base.BaseModel import BaseModel, ListType

from kindenty.model.DatabaseModel import Account


class AccountTradeLogDto(BaseModel):
    __slots__ = (
        'account', 'accountType', 'id', 'strategyId', 'type', 'status', 'price', 'power', 'avgPrice', 'orderCount',
        'time',
        'tradeTime', 'orderNo', 'buyCount', 'buyId')


class PriceDto(BaseModel):
    __slots__ = ('price', 'power')


class OrderBaseDto(BaseModel):
    __slots__ = ('buyPrice', 'buyCount', 'power', 'logId')


class SharesDto(BaseModel):
    __slots__ = (
        'fallX', 'buyN', 'growthRatio', 'buyM', 'buyCountQ', 'riseY', 'incrementA', 'sinkDayW', 'price', 'spRatio',
        'spTimes', 'intervalRatio', 'maxSell', 'volatility', 'sharesCode', 'intervalRatioX', 'intervalRatioY',
        'finalPrice', 'priceInterval', 'correctRatioN')


class SharesDto130(BaseModel):
    __slots__ = (
        'fallX', 'buyN', 'growthRatio', 'buyM', 'riseY', 'maxPrice', 'curPrice', 'sharesCode', 'maxSell')


class SharesDto140(BaseModel):
    __slots__ = (
        'riseY', 'sellN', 'sharesCode', 'price', 'total', 'curPower', 'growthRatio', 'maxSell')


class StrategyDto(BaseModel):
    __slots__ = ('version', 'operation', 'accounts', 'strategies', 'name', 'sharesCodes', 'sellStatus')

    accounts = ListType(Account)
    strategies = ListType(SharesDto)


class SubmitOrderDto(BaseModel):
    __slots__ = ('accountType', 'account', 'sharesCode', 'tradeSide', 'priceType', 'priceValue', 'buyCount', 'logId')


class CancelOrderDto(BaseModel):
    __slots__ = ('accountType', 'account', 'orderNo', 'logId')


class RealMarketData(BaseModel):
    __slots__ = (
        'stopMark', 'symbol', 'highPrice', 'lastPrice', 'time', 'perAssurescaleValue', 'openPrice', 'preClosePrice',
        'assureEnbuyBalance', 'upperLimit', 'lowerLimit', 'change', 'bailBalance')


class YieldDataDto(BaseModel):
    __slots__ = ('sharesCode', 'buyAmount', 'sellAmount')


class AssistDto(BaseModel):
    __slots__ = ('sharesCode', 'balance', 'enableCount', 'buyAmount', 'buyCount', 'sellAmount', 'sellCount')


class UserInvestmentDto(BaseModel):
    __slots__ = ('id', 'time', 'userName', 'amount', 'shares', 'netValue')


class InvestmentYeildDetailDto(BaseModel):
    __slots__ = ('time', 'amount', 'shares', 'buyNetValue', 'yieldRate', 'profit')


class UserAccountInfo(BaseModel):
    __slots__ = ('userName', 'totalAssets', 'totalShares', 'totalProfit', 'yieldRate')


class SellStatusDto(BaseModel):
    __slots__ = ('sharesCode', 'startCount', 'stopCount', 'version')



import threading
from queue import Queue
from abc import abstractmethod

from kindenty.base.DatabaseType import DatabaseType


class PoolException(Exception):
    pass


class Pool(object):
    """一个数据库连接池"""
    localConn = threading.local()
    localConn.conn = None
    def __init__(self, maxActive=5, maxWait=None, init_size=0, db_type: DatabaseType = DatabaseType.sqlite, **config):
        self.__freeConns = Queue(maxActive)
        self.maxWait = maxWait
        self.db_type = db_type
        self.config = config
        if init_size > maxActive:
            init_size = maxActive
        for i in range(init_size):
            self.free(self._create_conn())

    def __del__(self):
        self.release()

    def release(self):
        """释放资源，关闭池中的所有连接"""
        while self.__freeConns and not self.__freeConns.empty():
            con = self.get()
            con.release()
        self.__freeConns = None

    def _create_conn(self):
        """创建连接 """
        return supportDriver[self.db_type.name](**self.config)

    def get(self, timeout=None):
        """获取一个连接
        @param timeout:超时时间
        """
        if hasattr(self.localConn, 'conn') and self.localConn.conn:
            return self.localConn.conn
        if timeout is None:
            timeout = self.maxWait
        conn = None
        if self.__freeConns.empty():  # 如果容器是空的，直接创建一个连接
            conn = self._create_conn()
        else:
            conn = self.__freeConns.get(timeout=timeout)
        conn.pool = self
        self.localConn.conn = conn
        return conn

    def free(self, conn):
        """将一个连接放回池中
        @param conn: 连接对象
        """
        conn.pool = None
        if self.__freeConns.full():  # 如果当前连接池已满，直接关闭连接
            conn.release()
            return
        self.__freeConns.put_nowait(conn)


class PoolingConnection(object):
    def __init__(self, **config):
        self.conn = None
        self.config = config
        self.pool = None

    def __del__(self):
        self.release()

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def release(self):
        if self.conn is not None:
            self.conn.close()
            self.conn = None
        self.pool = None

    def close(self):
        if self.pool is None:
            raise PoolException("连接已关闭")
        self.pool.free(self)

    def __getattr__(self, val):
        if self.conn is None and self.pool is not None:
            self.conn = self._create_conn(**self.config)
        if self.conn is None:
            raise PoolException("无法创建数据库连接 或连接已关闭")
        return getattr(self.conn, val)

    @abstractmethod
    def _create_conn(self, **config):
        pass


class SQLit3PoolConnection(PoolingConnection):
    def _create_conn(self, **config):
        import sqlite3
        return sqlite3.connect(**config)


class MysqlPoolConnection(PoolingConnection):
    def _create_conn(self, **config):
        import pymysql
        return pymysql.connect(**config)


supportDriver = {'sqlite': SQLit3PoolConnection, 'mysql': MysqlPoolConnection}



class ArgsException(Exception):

    def __init__(self, *args, **kw):
        super(ArgsException, self).__init__(*args, **kw)


class ModelException(Exception):

    def __init__(self, *args, **kw):
        super(ModelException, self).__init__(*args, **kw)


class ResultMultipleException(Exception):

    def __init__(self, *args, **kw):
        super(ModelException, self).__init__(*args, **kw)


class DataException(Exception):

    def __init__(self, *args, **kw):
        super(ModelException, self).__init__(*args, **kw)






import os
from configparser import ConfigParser

from kindenty.base import log

from kindenty.base.DatabaseType import DatabaseType
from kindenty.base.DatabasePool import Pool


# pool = Pool(database="/sqlit3Data.db", isolation_level=None)
# pool = Pool(check_same_thread=False)
# pool = Pool(db_type=DatabaseType.mysql, host='localhost', user='admin', password='123456', database='cats_strategy',
#             cursorclass=pymysql.cursors.DictCursor)
log.info('root Path: %s' % os.getcwd())
configParser = ConfigParser()
configParser.read('conf/config.ini')
host = configParser.get('database', 'host')
port = configParser.get('database', 'port')
user = configParser.get('database', 'userName')
password = configParser.get('database', 'password')
database = configParser.get('database', 'database')
driverName = configParser.get('database', 'driverName')
type = DatabaseType[driverName]

pool = Pool(maxActive=15, db_type=DatabaseType[driverName], host=host, user=user, password=password,
            database=database, autocommit=True, port=int(port))
